type GuiCondState = Maybe (Int,  Bool,   Bool, Int, Int)
--                  (while, update, history, cond, rule)

guiState :: AsmVar GuiCondState
guiState = initVal "guiState" Nothing

guiFunctionState :: AsmVar (Int -> Bool)
guiFunctionState = initAssocs' "guiDoNotShowFunction" asmDefault asmLt (==) []

instance AsmTerm GuiCondState

guiFireRule = rule
   where (Just (_,_,_,_,rule)) = guiState

class AsmShow a where
  asmShow :: a -> String
  asmShowAssocs :: a -> [(String,String)]


instance Text a => AsmShow a where
  asmShow = show
  asmShowAssocs = const []

instance Text [(a,b)] => AsmShow (a -> b) where
  asmShow = show . assocs
  asmShowAssocs = map (\(l,r) -> (show l, show r)) . assocs


guiOpenRuleBox :: GVar (Maybe Window) -> GVar (Rule ()) ->
     [(String,Rule ())] -> GUI ()
guiOpenRuleBox grules frule rules = do
   wrules <- readGVar grules
   case wrules of
    Nothing -> do 
      w <- window [title "Rules"]
      l <- (listbox [height 6, width 16] w :: GUI (Listbox [String]))
      if guiFireRule >= 0 then
        setSelection l [guiFireRule]
       else done
      v <- vscroll [takeFocus False] l
      b <- button [text "close", 
                   command (do closeWindow w
                               writeGVar grules Nothing)] w
      pack ((flexible l <|< v) ^-^ b)
      setValue l (map fst rules)
      cset l (on (click 1) 
                 (do [i] <- getSelection l
                     writeGVar frule (snd(rules!!i))
                     liftIO (let Just (w,u,h,c,_) = guiState
                             in  guiState !:= Just (w,u,h,c,i))))
      writeGVar grules (Just w)
      focus l
    Just w  ->  do 
      s <- tk_getTcl ["wm","state", tk_getPathName w]
      if s == "iconic"
       then tk_putTcl ["wm","deiconify", tk_getPathName w]
       else tk_putTcl ["wm","iconify", tk_getPathName w]  


guiOpenCondsBox :: GVar (Maybe Window) -> [(String, Bool)] ->
    GUI ()
guiOpenCondsBox gconds conds = do
   wconds <- readGVar gconds
   case wconds of
    Nothing -> do 
      w <- window [title "Conditions"]
      rwhile   <- radiobutton [text "while", anchor "w"] w
      runtil   <- radiobutton [text "until", anchor "w"] w
      let cmd i = liftIO (let Just (wu,u,h,c,r) = guiState
                          in  guiState !:= Just (i,u,h,c,r))
      cset rwhile (command (cmd 0))
      cset runtil (command (cmd 1))
      let Just (wu,u,h,c,r) = guiState
      radio [initValue wu] [rwhile, runtil]
      rtrue    <- radiobutton [text "True", anchor "w"] w
      rconds   <- binds [ radiobutton [ text c, anchor "w" ] w
                        | c <- map fst conds ]
      rc <- radio [initValue c] (rtrue : rconds)
      let cmd = do v <- getValue rc
                   liftIO (let Just (wu,u,h,c,r) = guiState
                           in  guiState !:= Just (wu,u,h,v,r))
      cset rtrue (command cmd)
      seqs [ cset c (command cmd) | c <- rconds ]
      cupdate  <- checkbutton [text "show updates", anchor "w",
                               initValue u] w
      chistory <- checkbutton [text "history", anchor "w", 
                               initValue h] w
      cset chistory (command (do
              v <- getValue chistory
              liftIO (let Just (wu,u,h,c,r) = guiState
                      in  do guiState !:= Just (wu,u,v,c,r)
                             primAsmSetHistory v
                      )))
      cset cupdate (command (do
              v <- getValue cupdate
              liftIO (let Just (wu,u,h,c,r) = guiState
                      in  guiState !:= Just (wu,v,h,c,r))))
      bclose   <- button [text "close",
                          command (do closeWindow w
                                      writeGVar gconds Nothing)] w
      fl <- frame [] (rwhile ^-^ runtil)
      fr <- frame [] (vertical (map fillX (rtrue : rconds)))
      pack ((fl << fr) ^-^  cupdate ^-^ chistory ^-^ bclose)
      writeGVar gconds (Just w)
      
    Just w  ->  do 
      s <- tk_getTcl ["wm","state", tk_getPathName w]
      if s == "iconic"
       then tk_putTcl ["wm","deiconify", tk_getPathName w]
       else tk_putTcl ["wm","iconify", tk_getPathName w]  

general_guiMain :: String -> Bool -> Bool ->
       String ->
       [(Int,String,String,[(String,String)])] ->
       [(String, Rule())] -> 
       [(String, Bool)] -> IO ()
general_guiMain titleStr 
                withHistory
                doUpdate
                fontStr
                guiDisplayFunctions 
                rules
                conds
     = start $ do
  gw <- newGVar ([] :: [(Int, (Window, GUI ()))])
  case guiState of
    Nothing -> liftIO (do 
                  let fireRule = if length rules > 0 then 0 else -1
                  guiState !:= Just (0,doUpdate,withHistory,0,fireRule)
                  primAsmSetHistory withHistory
                  seqs [ guiFunctionState i !:= True 
                       | i <- [1.. length guiDisplayFunctions - 1 ] ]
               )
    Just (wu,u,h,c,r) -> liftIO (primAsmSetHistory h)
  grules <- newGVar (Nothing :: Maybe Window)
  gconds <- newGVar (Nothing :: Maybe Window)
  frule  <- newGVar (if guiFireRule>=0 
                       then snd(rules!!guiFireRule) 
                       else skip)
  w <- window [title titleStr]
  bquit <- button [text "quit", command quit] w
  bfire <- button [text "fire"]  w
  brules<- button [text "rules", command (guiOpenRuleBox grules frule rules)] w
  brun  <- button [text "run"]   w
  bcond <- button [text "conds", command (guiOpenCondsBox gconds conds)] w
  badd  <- button [text "+"]     w
  bsub  <- button [text "-"]     w
  steps <- liftIO primAsmSteps
  liftIO (primAsmShowHistory steps)
  esteps <- entry [initValue steps, width 5] w
  let (_:funs) = guiDisplayFunctions
  lbox <- listbox [ height 4 , font fontStr] w
  let updFuns = do is <- getSelection lbox
                   setValue lbox (map (\(i,l,r,_) -> 
                     l ++ dynamicExpr (if guiFunctionState i 
                                         then r else "")) funs)
                   setSelection lbox is
  updFuns
  v <- vscroll [] lbox
  h <- hscroll [] lbox
  fbuttons <- frame [] 
     (bquit << brules << bcond <<
      brun << bsub << esteps << badd << bfire)
  pack (flexible (flexible (flexible (flexible lbox << fillY v) ^-^ fillX h)
         ^-^ fillX fbuttons))
  let update      = do updFuns
                       ws <- readGVar gw
                       seqs (map (snd.snd) ws)
                       updateTask
  let updateGUI   = do steps <- liftIO primAsmSteps
                       dynamicExpr (setValue esteps steps); update
  let cFire = do r <- readGVar frule
                 b <- liftIO (fire1_ (rule2io r))
                 b =>> updateGUI
  let activate b = do seqs [ cset widget (active b)
                           | widget <- [ bcond, bfire, brules,
                                         brun, badd, bsub ]
                           ]
  let cRun  = do r <- readGVar frule
                 let Just (wu,u,h,c,_) = guiState
                 if u
                  then case take 1 (drop c (("True",True) : conds)) of
                   [(_,c)] -> 
                     let loopw = do
                            done
                            if dynamicExpr c then do 
                               b <- liftIO (fire1_ (rule2io r))
                               if b then do 
                                 u =>> updateGUI
                                 after 0 loopw
                                else activate True
                             else activate True
                         loopu = do 
                            b <- liftIO (fire1_ (rule2io r))
                            if b then do
                              u =>> updateGUI
                              if not (dynamicExpr c) then after 0 loopu
                               else activate True
                             else activate True
                     in do activate False
                           if wu == 0 then do loopw; updateGUI
                                      else do loopu; updateGUI
                  else do
                   activate False
                   liftIO (do
                       if c == 0 then fixpoint r
                        else void $
                         case take 1 (drop (c-1) conds) of
                           [(_,c)] -> if wu == 0  then fireWhile c r
                                                  else fireUntil c r)
                   activate True
                   updateGUI
  cset bfire (command cFire)
  cset brun  (command cRun)
  cset lbox  (on (click 3) (do
      is <- getSelection lbox
      case is of 
       [i] -> do liftIO (guiFunctionState (i+1) !:= 
                              not(guiFunctionState (i+1)))
                 update
                 setSelection lbox [i]
                 tk_toTcl lbox ["see",show i]
       _   -> done))
  let createAssocsWindow i =
        case take 1 (drop i funs) of
          [(_,str,_,assocs)] -> do
                w <- window [title str]
                lbox <- listbox [font fontStr] w
                v <- vscroll [] lbox
                h <- hscroll [] lbox
                b <- button [text "close"] w
                cset b (command (do
                    ws <- readGVar gw
                    let ws' = filter ((/=i).fst) ws
                    writeGVar gw ws'
                    closeWindow w))
                pack (flexible (flexible (flexible lbox << fillY v) 
                                 ^^ fillX h)
                      ^^ fillX b)
                let fill n xs = take n (xs ++ repeat ' ') ++ "  "
                let upd = do
                        is <- getSelection lbox
                        dynamicExpr (
                           let n = maximum (0: map length (map fst assocs)) 
                           in  setValue lbox (map (\(l,r) -> fill n l ++ r) 
                                               assocs))
                        setSelection lbox is
                upd
                ws <- readGVar gw
                writeGVar gw ((i,(w,upd)):ws)
          _ -> error "createAssocsWindow"
  cset lbox (on (doubleClick 1) (do
         [i] <- getSelection lbox
         wl <- readGVar gw
         case filter ((==i).fst) wl of
          [(i,(w,_))] -> do 
              s <- tk_getTcl ["wm","state", tk_getPathName w]
              if s == "iconic"
                then tk_putTcl ["wm","deiconify", tk_getPathName w]
                else tk_putTcl ["wm","iconify", tk_getPathName w]  
          _  -> createAssocsWindow i
       ))
  let displayHistory = do v <- getValue esteps
                          b <- liftIO (primAsmShowHistory v)
                          if not b then do
                            steps <- liftIO primAsmSteps
                            setValue esteps steps
                           else done
                          update
  cset esteps (on return displayHistory)
  cset badd (command (do v <- getValue esteps
                         setValue esteps (v+1)
                         displayHistory))
  cset bsub (command (do v <- getValue esteps
                         m <- liftIO primAsmMinHistory
                         setValue esteps (max(v-1)m)
                         displayHistory))
guiDisplayFunctions :: [(Int,String,String,[(String,String)])]
guiDisplayFunctions = [ (0,[],[],[])
  , (1, "stack     ", asmShow (stack ),asmShowAssocs (stack ))
  , (2, "auxPole   ", asmShow (auxPole ),asmShowAssocs (auxPole ))
  , (3, "toPole    ", asmShow (toPole ),asmShowAssocs (toPole ))
  , (4, "fromPole  ", asmShow (fromPole ),asmShowAssocs (fromPole ))
  , (5, "goal      ", asmShow (goal ),asmShowAssocs (goal ))
  , (6, "call      ", asmShow (call ),asmShowAssocs (call ))
  ]

guiUsedRules :: [(String, Rule ())]
guiUsedRules = [("hanoi", hanoi)]

guiUsedConds :: [(String, Bool)]
guiUsedConds = [("False", False), ("call == 10", call == 10), ("call >= 0", call >= 0)]

guiMain :: IO ()
guiMain = general_guiMain "main"True True "7x14"  guiDisplayFunctions guiUsedRules guiUsedConds

